#include "Obelisk.h"

#include "CGameController.h"
#include "CMap.h"
#include "CKnifeBot.h"
#include "CBlade.h"
#include "CPowerUp.h"

using namespace std;
using namespace Obelisk;



CMap::CMap(IObelisk* aObelisk, CGameController* aGameController, int aGridXSize, int aGridYSize)
{
	mGameController = aGameController;

	mGridXSize = aGridXSize;
	mGridYSize = aGridYSize;
	mXSize = mGridXSize * 36;
	mYSize = mGridYSize * 36;

	mPlayerX = 50;
	mPlayerY = 50;
	mPlayerDir = 1;
	mPlayerNumLives = 2;
	mPlayerInvincibleTime = 0;
	mPlayerLaunchTime = 0;

	const int cNumKnifeBots = 16;

	for (int i = 0; i < cNumKnifeBots; i++)
	{
		int X;
		int Y;
		while (true)
		{
			X = aObelisk->Random(0, mGridXSize - 1);
			Y = aObelisk->Random(0, mGridYSize - 1);
			if (X >= 6 || Y >= 6)
				break;
		}
		int Dir = aObelisk->Random(0, 3);
		mKnifeBots.Add(new CKnifeBot(this, X, Y, Dir, aObelisk->Random(0, 1) != 0));
	}

	for (int i = 0; i < 180; i++)
	{
		int X = aObelisk->Random(10, mXSize - 21);
		int Y = aObelisk->Random(10, mYSize - 21);
		int Angle = aObelisk->Random(0, 359);
		mBlades.Add(new CBlade(this, X, Y, Angle));
	}

	mKnifeBotBoundBox.push_back(pair<int, int>(-17, -17));
	mKnifeBotBoundBox.push_back(pair<int, int>(17, -17));
	mKnifeBotBoundBox.push_back(pair<int, int>(17, 17));
	mKnifeBotBoundBox.push_back(pair<int, int>(-17, 17));
	mKnifeBotBoundBoxBox = pair<int, int>(51, 51);

	mBladeBoundBox.push_back(pair<int, int>(-41, -11));
	mBladeBoundBox.push_back(pair<int, int>(41, -12));
	mBladeBoundBox.push_back(pair<int, int>(30, -3));
	mBladeBoundBox.push_back(pair<int, int>(8, 8));
	mBladeBoundBox.push_back(pair<int, int>(-11, 10));
	mBladeBoundBox.push_back(pair<int, int>(-41, 12));
	mBladeBoundBoxBox = pair<int, int>(106, 106);

	mPlayerBoundBox.push_back(pair<int, int>(-3, -12));
	mPlayerBoundBox.push_back(pair<int, int>(3, -12));
	mPlayerBoundBox.push_back(pair<int, int>(3, 0));
	mPlayerBoundBox.push_back(pair<int, int>(-3, 0));
	mPlayerBoundBoxBox = pair<int, int>(6, 12);

	mPowerUpBoundBox.push_back(pair<int, int>(-6, -9));
	mPowerUpBoundBox.push_back(pair<int, int>(6, -9));
	mPowerUpBoundBox.push_back(pair<int, int>(10, 0));
	mPowerUpBoundBox.push_back(pair<int, int>(6, 9));
	mPowerUpBoundBox.push_back(pair<int, int>(-6, 9));
	mPowerUpBoundBox.push_back(pair<int, int>(-10, 0));
	mPowerUpBoundBoxBox = pair<int, int>(20, 18);

	mRobotSpawnTime = 50 * 5;

	mPowerUpTime = 30 * 5;
}



CMap::~CMap()
{
	foreach (mKnifeBots)
		delete mKnifeBots.Current();
	foreach (mBlades)
		delete mBlades.Current();
	foreach (mPowerUps)
		delete mPowerUps.Current();
}



void CMap::MovePlayer(int aXDir, int aYDir)
{
	const int cMove = 2;
	if (aYDir != 0)
	{
		mPlayerY += aYDir * cMove;
		mPlayerDir = aYDir == -1 ? 3 : 1;
	}
	if (aXDir != 0)
	{
		mPlayerX += aXDir * cMove;
		mPlayerDir = aXDir == -1 ? 2 : 0;
	}
	if (mPlayerX < 0)
		mPlayerX = 0;
	else if (mPlayerX >= mXSize)
		mPlayerX = mXSize - 1;
	if (mPlayerY < 0)
		mPlayerY = 0;
	else if (mPlayerY >= mYSize)
		mPlayerY = mYSize - 1;
}



void CMap::Step(IObelisk* aObelisk)
{
	if (mPlayerInvincibleTime != 0)
		mPlayerInvincibleTime--;
	if (mPlayerLaunchTime != 0)
	{
		mPlayerLaunchTime--;
		double Angle = mPlayerVelocity.first * M_PI / 180;
		mPlayerRealPos.first += cos(Angle) * mPlayerVelocity.second;
		mPlayerRealPos.second += sin(Angle) * mPlayerVelocity.second;
		mPlayerVelocity.second *= .92;
		if (mPlayerRealPos.first < 0 && mPlayerVelocity.first >= 90 && mPlayerVelocity.first <= 270)
		{
			mPlayerRealPos.first = -mPlayerRealPos.first;
			mPlayerVelocity.first = 180 - mPlayerVelocity.first;
			if (mPlayerVelocity.first < 0)
				mPlayerVelocity.first += 360;
		}
		if (mPlayerRealPos.first >= mXSize && (mPlayerVelocity.first >= 270 || mPlayerVelocity.first <= 90))
		{
			mPlayerRealPos.first = GetXSize() - (mPlayerRealPos.first - GetXSize());
			mPlayerVelocity.first = 180 - mPlayerVelocity.first;
			if (mPlayerVelocity.first < 0)
				mPlayerVelocity.first += 360;
		}
		if (mPlayerRealPos.second < 0 && mPlayerVelocity.first >= 180)
		{
			mPlayerRealPos.second = -mPlayerRealPos.second;
			mPlayerVelocity.first = 360 - mPlayerVelocity.first;
		}
		if (mPlayerRealPos.second >= GetYSize() && mPlayerVelocity.first <= 180)
		{
			mPlayerRealPos.second = GetYSize() - (mPlayerRealPos.second - GetYSize());
			mPlayerVelocity.first = 360 - mPlayerVelocity.first;
		}
		mPlayerX = mPlayerRealPos.first;
		mPlayerY = mPlayerRealPos.second;
	}

	if (--mRobotSpawnTime == 0)
	{
		mRobotSpawnTime = 5 * 50;
		int X;
		int Y;
		while (true)
		{
			X = aObelisk->Random(0, mGridXSize - 1);
			Y = aObelisk->Random(0, mGridYSize - 1);
			if (abs(X * 36 + 18 - mPlayerX) >= 36 * 6 || abs(Y * 36 + 18 - mPlayerY) >= 36 * 6)
				break;
		}
		int Dir = aObelisk->Random(0, 3);
		mKnifeBots.Add(new CKnifeBot(this, X, Y, Dir, aObelisk->Random(0, 1) != 0));
	}

	if (--mPowerUpTime == 0)
	{
		mPowerUpTime = 30 * 5;
		if (mPowerUps.GetCount() < 12)
		{
			int X;
			int Y;
			while (true)
			{
				X = aObelisk->Random(10, mXSize - 11);
				Y = aObelisk->Random(10, mYSize - 11);
				if (abs(X - mPlayerX) >= 36 * 6 || abs(Y - mPlayerY) >= 36 * 6)
					break;
			}
			mPowerUps.Add(new CPowerUp(X, Y, aObelisk->Random(0, 1)));
		}
	}
}



void CMap::LaunchPlayer(Obelisk::IObelisk* aObelisk, float aAngle, bool aFast)
{
	mPlayerLaunchTime = 60;
	int Velocity = aObelisk->Random(aFast ? 25 : 10, aFast ? 30 : 20);
	if (aAngle > 360)
		aAngle -= 360;
	mPlayerVelocity = pair<float, float>(aAngle, Velocity);
	mPlayerRealPos = pair<float, float>(mPlayerX, mPlayerY);
}



void CMap::CheckCollisions(IObelisk* aObelisk)
{
	foreach (mPowerUps)
	{
		if (Collision(mPlayerBoundBox, mPlayerBoundBoxBox, mPlayerX, mPlayerY, 0,
				mPowerUpBoundBox, mPowerUpBoundBoxBox, mPowerUps->GetX(), mPowerUps->GetY(), 0))
		{
			if (mPowerUps->GetType() == 0)
				mPlayerNumLives++;
			else
				mGameController->GetEMP();
			delete mPowerUps.Current();
			mPowerUps.Delete();
		}
	}

	foreach (mBlades)
	{
		if (mBlades->IsActive())
		{
			if (mPlayerInvincibleTime == 0)
			{
				if (Collision(mBladeBoundBox, mBladeBoundBoxBox, mBlades->GetX(), mBlades->GetY(), mBlades->GetAngle(),
						mPlayerBoundBox, mPlayerBoundBoxBox, mPlayerX, mPlayerY, 0))
				{
					mPlayerNumLives--;
					mPlayerInvincibleTime = 3 * 50;
					LaunchPlayer(aObelisk, mBlades->IsAttached() ? mBlades->GetAngle() + 90:
							aObelisk->Random(0, 359), true);
				}
			}

			CCollection<CBlade> Blades2 = mBlades;
			bool Stop = false;
			foreach (Blades2)
			{
				if (mBlades.Current() == Blades2.Current() || !Blades2->IsAttached() ||
						mBlades->GetKnifeBot() == Blades2->GetKnifeBot())
					continue;
				if (Collision(mBladeBoundBox, mBladeBoundBoxBox, mBlades->GetX(),
						mBlades->GetY(), mBlades->GetAngle(), mBladeBoundBox,
						mBladeBoundBoxBox, Blades2->GetX(), Blades2->GetY(), Blades2->GetAngle()))
				{
					if (mBlades->GetGeneration() != -1)
					{
						Blades2->SetGeneration(mBlades->GetGeneration() + 1);
						mGameController->BladeScore(mBlades->GetGeneration() + 1);
					}
					Blades2->Launch(aObelisk);
					if (mBlades->IsAttached())
						mBlades->Launch(aObelisk);
					if (DistanceBetween(mBlades->GetX(), mBlades->GetY(), mPlayerX, mPlayerY) < 600)
						mGameController->BladeCrash();
					Stop = true;
					break;
				}
			}
			if (Stop)
				continue;
		}
		foreach (mKnifeBots)
		{
			if (mBlades->GetKnifeBot() == mKnifeBots.Current())
				continue;
			if (Collision(mKnifeBotBoundBox, mKnifeBotBoundBoxBox, mKnifeBots->GetX(), mKnifeBots->GetY(),
					mKnifeBots->GetAngle(), mBladeBoundBox, mBladeBoundBoxBox, mBlades->GetX(), mBlades->GetY(),
					mBlades->GetAngle()))
			{
				if (mBlades->IsStationary())
				{
					if (mKnifeBots->NumBlades() < 4)
						mKnifeBots->AttachBlade(aObelisk, mBlades.Current());
				}
			}
		}
	}

	if (mPlayerInvincibleTime == 0)
	{
		foreach (mKnifeBots)
		{
			if (Collision(mKnifeBotBoundBox, mKnifeBotBoundBoxBox, mKnifeBots->GetX(), mKnifeBots->GetY(), mKnifeBots->GetAngle(),
					mPlayerBoundBox, mPlayerBoundBoxBox, mPlayerX, mPlayerY, 0))
			{
				mPlayerNumLives--;
				mPlayerInvincibleTime = 3 * 50;
				LaunchPlayer(aObelisk, aObelisk->Random(0, 359), false);
				break;
			}
		}
	}
}



bool CMap::Collision(const vector<pair<int, int>>& aBoundBox0, const pair<int, int>& aBoundBoxBox0, int aX0, int aY0, int aAngle0,
		const vector<pair<int, int>>& aBoundBox1, const pair<int, int>& aBoundBoxBox1, int aX1, int aY1, int aAngle1)
{
	int BB0X0 = aX0 - aBoundBoxBox0.first / 2 - 1;
	int BB0X1 = aX0 + aBoundBoxBox0.first / 2 - 1;
	int BB0Y0 = aY0 - aBoundBoxBox0.second / 2 - 1;
	int BB0Y1 = aY0 + aBoundBoxBox0.second / 2 - 1;
	int BB1X0 = aX1 - aBoundBoxBox1.first / 2 - 1;
	int BB1X1 = aX1 + aBoundBoxBox1.first / 2 - 1;
	int BB1Y0 = aY1 - aBoundBoxBox1.second / 2 - 1;
	int BB1Y1 = aY1 + aBoundBoxBox1.second / 2 - 1;
	if (BB0X0 > BB1X1 || BB0X1 < BB0X0 || BB0Y0 > BB1Y1 || BB0Y1 < BB1Y0)
		return false;
	return Collision1(aBoundBox0, aX0, aY0, aAngle0, aBoundBox1, aX1, aY1, aAngle1) ||
			Collision1(aBoundBox1, aX1, aY1, aAngle1, aBoundBox0, aX0, aY0, aAngle0);
}



bool CMap::Collision1(const vector<pair<int, int>>& aBoundBox0, int aX0, int aY0, int aAngle0,
		const vector<pair<int, int>>& aBoundBox1, int aX1, int aY1, int aAngle1)
{
	for (int i = 0; i < aBoundBox0.size(); i++)
	{
		pair<float, float> p = TransformPoint(aBoundBox0[i], aX0, aY0, aAngle0);
		bool Inside = true;
		for (int j = 0; j < aBoundBox1.size(); j++)
		{
			int j1 = (j + 1) % aBoundBox1.size();
			pair<float, float> q0 = TransformPoint(aBoundBox1[j], aX1, aY1, aAngle1);
			pair<float, float> q1 = TransformPoint(aBoundBox1[j1], aX1, aY1, aAngle1);

			pair<float, float> vp = GetVector(q0, p);
			pair<float, float> v = GetVector(q0, q1);
			pair<float, float> vq = pair<float, float>(-v.second, v.first);   // rotate v by 90 degrees
			if (vp.first * vq.first + vp.second * vq.second < 0)
			{
				Inside = false;
				break;
			}
		}
		if (Inside)
			return true;
	}
	return false;
}



pair<float, float> CMap::TransformPoint(const pair<int, int>& p, int aX, int aY, float aAngle)
{
	pair<float, float> Result;
	double Angle = (double)aAngle * M_PI / 180;
	Result.first = p.first * cos(Angle) - p.second * sin(Angle) + aX;
	Result.second = p.second * cos(Angle) + p.first * sin(Angle) + aY;
	return Result;
}



pair<float, float> CMap::GetVector(const pair<float, float>& p0, const pair<float, float>& p1)
{
	return pair<float, float>(p1.first - p0.first, p1.second - p0.second);
}



void CMap::DetonateEMP(IObelisk* aObelisk, int aX, int aY)
{
	foreach (mBlades)
	{
		if (mBlades->IsFlying() && DistanceBetween(mBlades->GetX(), mBlades->GetY(), aX, aY) < 250)
		{
			float Angle = atan2((double)mBlades->GetY() - aY, (double)mBlades->GetX() - aX) * 180 / M_PI;
			if (Angle < 0)
				Angle += 360;
			float OppositeAngle = Angle + 180;
			if (OppositeAngle >= 360)
				OppositeAngle -= 360;
			float FlyingAngle = mBlades->GetFlyingAngle();
			float Diff = abs(FlyingAngle - OppositeAngle);
			if (Diff > 180)
				Diff = 360 - Diff;
			if (Diff < 60)
			{
				mBlades->LaunchAngle(aObelisk, Angle + aObelisk->Random(0, 6) - 3);
				mBlades->SetGeneration(0);
			}
		}
	}
	foreach (mKnifeBots)
	{
		if (pow((double)abs(mKnifeBots->GetX() - aX), 2) + pow((double)abs(mKnifeBots->GetY() - aY), 2) < 280 * 280)
		{
			for (int i = mKnifeBots->NumBlades() - 1; i >= 0; i--)
			{
				float Angle = atan2((double)mKnifeBots->GetY() - aY, (double)mKnifeBots->GetX() - aX) * 180 / M_PI;
				if (Angle < 0)
					Angle += 360;
				mGameController->BladeScore(1);
				mKnifeBots->GetBlade(i)->SetGeneration(0);
				mKnifeBots->GetBlade(i)->LaunchAngle(aObelisk, Angle + aObelisk->Random(0, 6) - 3);
			}
		}
	}
}



int CMap::DistanceBetween(int aX0, int aY0, int aX1, int aY1)
{
	return sqrt(pow((double)aX1 - aX0, 2) + pow((double)aY1 - aY0, 2));
}